#ifndef AIRPORTSIM_H
#define AIRPORTSIM_H

#include <vector>
#include <deque>
#include <string>
#include <queue>

#define MAX_INT 32768
extern float minCheckinTime;
const float windowSize = 512;
const float windowScale = 1.0f/windowSize;
extern float bodyWidth;

class Distribution {
public:
	virtual double getNumber() = 0;
};

class LinearDist {
public:
	float low, high;
	LinearDist(float p_low, float p_high) : low(p_low), high(p_high) {}
	double getNumber();
};

/* Jay's modifications */
class WaitDist : Distribution {
public:
	float min_, max_, peak_, range_, ret_range_l_, ret_range_h_;
	int power_;
	WaitDist() {}
	WaitDist(float min, float max, float peak, int power);
	virtual double getNumber();
};


class SpeedDist : public Distribution {
	public:
	static double speed_dist_lookup[51];
	virtual double getNumber();
};

/* end Jay's modifications */

class ArrivalDist : public Distribution {
public:
	static double arrival_dist_lookup[41];
	virtual double getNumber();
	virtual double max();
};

extern double complaint(double wait_time);


extern ArrivalDist arrivalDist;
extern WaitDist checkinWait; extern WaitDist securityWait; extern WaitDist bathroomWait;
 extern SpeedDist baseSpeedDistribution; extern SpeedDist disabledBaseSpeedDistribution;  extern SpeedDist escortBaseSpeedDistribution;
extern WaitDist timeTillNextFlight;

	class Flight {
		public:
		float minutes;  //  minutes left until flight
		int number;
		int gate;
		int capacity;
		
		Flight() {}
		Flight(float p_minutes, int p_number, int p_gate, int p_capacity) : minutes(p_minutes), number(p_number), gate(p_gate), capacity(p_capacity) {}
	};

class Graph {
	public:
	enum { FREE=0, SECURITY_IN, SECURITY_OUT };
	
	class Vertex {
		public:
			std::vector<int> adj;
			std::vector<int> weight;
	};
			
	std::vector<Vertex> verts;
	void addVertex();
	void addEdge(int a, int b, int weight=FREE);
	void addEdges(int a, int low, int high);
	void addBiEdge(int a, int b);
	std::vector<int>& adjacent(int i);
	std::vector<int>& weight(int i);
	int totalVertices();
};

class WaitProcess {
	public:
		class WaitLine {
			public:
			std::deque<int> waiters;
			float timeLeft;
		};
		std::vector<WaitLine> lines;
		std::vector<int> doneWaiting;
		WaitDist numDist;
		
		WaitProcess() {}
		void think(double tm);
		bool isDone(int gid);
		void add(int gid);
		void remove(int gid);
};

class Area {
	public:
		int numPeople;
		int maxPeople;
		float distance;
		float x, y;
		std::string name;
		int type;
		WaitProcess waitProcess;
		
		enum { FREE=0, START, WC, CHECKIN, SECURITY_IN, BATHROOM, RECREATION, BAGGAGE, GATE1 };
		
		Area(int type, float size, float x, float y);
		float congestion();
		float speedModifier();
		bool enter();
		void leave();
		void draw();
		void think(double tm);
		bool mustWait(int theid, WaitProcess * &wp);
};

class Wheelchair {
	public:
		bool motorized;
		float baseSpeed;
		float distance;
		int area;
		float timeInUse;
		
		Wheelchair() : timeInUse(0) {}
		Wheelchair(bool p_motorized);
		void resetStats();
};

class WheelchairBank {
	public:
		std::deque<Wheelchair> WC;
		std::deque<Wheelchair> motorWC;
		int area;
		
		WheelchairBank(int p_area);
		Wheelchair takeOne(bool motorized);
		void returnOne(Wheelchair wc);
		bool has(bool motorized);
		int total();
		void resetStats();
};

class Person {
	protected:
		int computeVisitList(bool byType, int start, int destinationType);
		int getEdgeWeight(int type);
	public:
		float baseSpeed;
		int curArea;
		int startArea;
		std::vector<int> toVisit;
		float distance;
		int destination;
		
		Person();
		void movement(double tm);
		virtual void extraEnter();
};

class Passenger : public Person {
	protected:
		int getEdgeWeight(int type);
		
	public:
		enum { SPAWNED=0, TO_CHECK_IN, TO_GATE, TO_BATHROOM, ARRIVED, TO_BAGGAGE, LEAVING };
		
		int state;
		float waitTime;
		WaitProcess *waitProcess;
		int myID;
		int flightID;
		int destGate;
		bool limbo;
		float x, y;
		bool hasWheelchair;
		Wheelchair wheelchair;
		bool waiting;
		float timeToSpawn;
		float timeSpawned;
		float timeEscorted;
		
		Passenger(int p_myID, Flight &flight, int p_startArea);
		virtual void handleState(double tm);
		void think(double tm);
		virtual void draw();
		virtual void die();
		void leave();
		virtual void debugOutput();
		void extraEnter();
		virtual void resetStats() {timeEscorted=0;waitTime=0;}
};

class DisabledPassenger : public Passenger {
	public:
		enum { NEED_ESCORT=Passenger::LEAVING+1, WAITING_FOR_ESCORT, INDEPENDENT };
		float totalEscortWait;
		int escortID;
		int type;
		bool needsEscort;
		bool wantsMotorized;
		std::vector<Wheelchair> wheelchair;
		
		DisabledPassenger(int p_myID, Flight &flight, int p_startArea);
		void handleState(double tm);
		int pickedUpByEscort(int p_escortID);	//  returns id of area they want to go
		void droppedOffByEscort(int area);
		void debugOutput();
		void draw();
		void die();
		void receiveWheelchair(Wheelchair wc);
		void dumpWheelchair();
		void resetStats();
};

class Escort : public Person {
	public:
		enum { SPAWNED=0, WAIT_FOR_CALL, TO_DISABLED, TO_DESTINATION, RETURN_WC, GATHER_WCS };
		int state;
		int myID;
		bool hasDisabled;
		int disabledID;
		float cash;
		float timeSinceLastPayment;
		float wages;
		float x,y;
		std::vector<Wheelchair> wheelchairs;
		float collectorTime;
		int lastState;
		float busyTime;
		
		Escort(int p_myID, int p_startArea, float p_wages);
		void think(double tm);
		void draw();
		void debugOutput();
		void resetStats();
};


class Airport {
	private:
		void spawnPassengers(int area, Flight &flight, int number);
		void spawnPassengersArrival(int area, int gate, int number);
		void removePassengers(int area, int flightNumber);
		int gateArea(int which);
		int areaGate(int which);
		int entrance();
		
		std::vector<int> gateAreas;
	public:
	
	std::vector<Flight> departures;
	std::vector<Flight> arrivals;
	
	Graph graph;
	std::vector<Area> areas;
	std::vector<Escort> escorts;
	std::vector<Passenger*> passengers;
	std::vector<WheelchairBank> wheelchairBanks;
	std::deque<Wheelchair> floatingWheelchairs;
	int gFlightID;
	int gPassengerID;
	int gEscortID;
	float lastWriteout;
		
	Airport();
	float die();
	void init();
	void step(double tm);
	void draw();
	void debugOutput();
	void printFlights();
	Passenger* passenger(int pid);
	Flight departure(int which);
	WheelchairBank& wcBank(int area);
	void outputStatistics(std::ofstream &out, float totalTime);
	void readParameters();
	void tweakParameters();
	void writeParameters();
};

#endif