/************************************************************************************
All Code by Troy Lawlor Copyright (c)2005 All Rights Reserved.
*************************************************************************************/

#include "3DMath.h"
#include <algorithm>

/*	MVector	*/

MVector::MVector() { v[0]=v[1]=v[2]=0; }

MVector::MVector( float p_x, float p_y, float p_z ) {
	v[0]=p_x; v[1]=p_y; v[2]=p_z;
}

MVector::MVector( float pv[] ) {
	v[0]=pv[0]; v[1]=pv[1]; v[2]=pv[2];
}

MVector MVector::normal() const {
	MVector r(*this);
	r.normalize();
	return r;
}

float MVector::length() const {
	return sqrt( (v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]) );
}

void MVector::normalize() {
	float magnitude=sqrt( (v[0]*v[0]) + (v[1]*v[1]) + (v[2]*v[2]) );
	v[0] /= magnitude; v[1] /= magnitude; v[2] /= magnitude;
}

MVector MVector::operator +(MVector rh) const {
	return MVector( v[0]+rh.v[0], v[1]+rh.v[1], v[2]+rh.v[2] );
}

MVector MVector::operator -(MVector rh) const {
	return MVector( v[0]-rh.v[0], v[1]-rh.v[1], v[2]-rh.v[2] );
}

MVector MVector::operator -() const {
	return MVector( -v[0], -v[1], -v[2] );
}

MVector MVector::operator *(float s) const {
	return MVector( v[0]*s, v[1]*s, v[2]*s );
}

float MVector::operator *(MVector rh) const {
	return dotProduct(*this, rh);
}

MVector MVector::operator *=(float s) {
	*this = *this * s;
	return *this;
}

MVector MVector::operator +=(MVector rh) {
	*this = *this + rh;
	return *this;
}

MVector MVector::operator -=(MVector rh) {
	*this = *this - rh;
	return *this;
}

float& MVector::operator[](int index) {
	return v[index];
}

const float& MVector::operator[](int index) const {
	return v[index];
}

float& MVector::x() { return v[0]; }
const float& MVector::x() const { return v[0]; }
float& MVector::y() { return v[1]; }
const float& MVector::y() const { return v[1]; }
float& MVector::z() { return v[2]; }
const float& MVector::z() const { return v[2]; }

float dotProduct( MVector a, MVector b )
{
	return a.v[0]*b.v[0] + a.v[1]*b.v[1] + a.v[2]*b.v[2];
}

MVector crossProduct( MVector a, MVector b )
{
	return MVector((a.v[1]*b.v[2] - a.v[2]*b.v[1]), (a.v[2]*b.v[0] - a.v[0]*b.v[2]), (a.v[0]*b.v[1] - a.v[1]*b.v[0]));
}

//  note: expects unit vectors
float angleBetween( MVector a, MVector b ) {
	return acos(dotProduct(a, b));
}

/*	MPlane	*/

MPlane::MPlane() : d(0) {}

MPlane::MPlane( MVector p_n, float p_d ) : n(p_n), d(p_d) {}

MPlane::MPlane( float p_a, float p_b, float p_c, float p_d ) : n(p_a, p_b, p_c), d(p_d) {}

void MPlane::normalize() {
	float magnitude=sqrt( (n[0]*n[0]) + (n[1]*n[1]) + (n[2]*n[2]) );
	n[0] /= magnitude; n[1] /= magnitude; n[2] /= magnitude; d /= magnitude;
}

float MPlane::operator *(MPlane p) const {
	return n*p.n + d*p.d;
}
MPlane MPlane::operator *(float s) const {
	return MPlane(n*s, d*s);
}

bool planeIntersection( const MPlane &a, const MPlane &b, const MPlane &c, MVector &result ) {
	float fDet;
	float MN[9] = { a.n.x(), a.n.y(), a.n.z(), b.n.x(), b.n.y(), b.n.z(), c.n.x(), c.n.y(), c.n.z() };
	float IMN[9] = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f };
	float MD[3] = { a.d, b.d , c.d };

	IMN[0] = MN[4] * MN[8] - MN[5] * MN[7];
	IMN[3] = -(MN[3] * MN[8] - MN[5] * MN[6]);
	IMN[6] = MN[3] * MN[7] - MN[4] * MN[6];

	fDet = MN[0] * IMN[0] + MN[1] * IMN[3] + MN[2] *IMN[6];

	//  no intersection
	if(fDet == 0.0f)
		return false;

	IMN[1] = -(MN[1] * MN[8] - MN[2] * MN[7]);
	IMN[4] = MN[0] * MN[8] - MN[2] * MN[6];
	IMN[7] = -(MN[0] * MN[7] - MN[1] * MN[6]);
	IMN[2] = MN[1] * MN[5] - MN[2] * MN[4];
	IMN[5] = -(MN[0] * MN[5] - MN[2] * MN[3]);
	IMN[8] = MN[0] * MN[4] - MN[1] * MN[3];

	fDet = 1.0f / fDet;

	IMN[0] *= fDet;
	IMN[1] *= fDet;
	IMN[2] *= fDet;
	IMN[3] *= fDet;
	IMN[4] *= fDet;
	IMN[5] *= fDet;
	IMN[6] *= fDet;
	IMN[7] *= fDet;
	IMN[8] *= fDet;

	result.x() = IMN[0] * MD[0] + IMN[1] * MD[1] + IMN[2] * MD[2];
	result.y() = IMN[3] * MD[0] + IMN[4] * MD[1] + IMN[5] * MD[2];
	result.z() = IMN[6] * MD[0] + IMN[7] * MD[1] + IMN[8] * MD[2];

	return true;
}

/*	MMatrix4	*/

//  initializes to the identity matrix
MMatrix4::MMatrix4() {
	for(int i=0; i<16; i++)
		e[i]=0;

	e[0] = 1;  e[5]=1;  e[10]=1;  e[15]=1;
}

MMatrix4::MMatrix4(float x, float y, float z, float xr, float yr, float zr) {
	*this = eulerToMatrix(xr,yr,zr);
	setTranslation(MVector(x,y,z));
}

MMatrix4::MMatrix4(float p_e[]) {
	std::copy(p_e, p_e+16, e);
}

MMatrix4::MMatrix4(MVector p) {
	MMatrix4 r;
	e[0] = p[0];  e[5]=p[1];  e[10]=p[2];
}

//  "post-multiply"
MMatrix4 MMatrix4::operator*( const MMatrix4& rh ) const {
	MMatrix4 m;

	/*m.e[0] = e[0]*rh.e[0] + e[4]*rh.e[1] + e[8]*rh.e[2];
	m.e[1] = e[1]*rh.e[0] + e[5]*rh.e[1] + e[9]*rh.e[2];
	m.e[2] = e[2]*rh.e[0] + e[6]*rh.e[1] + e[10]*rh.e[2];
	m.e[3] = 0;

	m.e[4] = e[0]*rh.e[4] + e[4]*rh.e[5] + e[8]*rh.e[6];
	m.e[5] = e[1]*rh.e[4] + e[5]*rh.e[5] + e[9]*rh.e[6];
	m.e[6] = e[2]*rh.e[4] + e[6]*rh.e[5] + e[10]*rh.e[6];
	m.e[7] = 0;

	m.e[8] = e[0]*rh.e[8] + e[4]*rh.e[9] + e[8]*rh.e[10];
	m.e[9] = e[1]*rh.e[8] + e[5]*rh.e[9] + e[9]*rh.e[10];
	m.e[10] = e[2]*rh.e[8] + e[6]*rh.e[9] + e[10]*rh.e[10];
	m.e[11] = 0;

	m.e[12] = e[0]*rh.e[12] + e[4]*rh.e[13] + e[8]*rh.e[14] + e[12];
	m.e[13] = e[1]*rh.e[12] + e[5]*rh.e[13] + e[9]*rh.e[14] + e[13];
	m.e[14] = e[2]*rh.e[12] + e[6]*rh.e[13] + e[10]*rh.e[14] + e[14];
	m.e[15] = 1;*/
	
	m.e[ 0] = rh.e[ 0] * e[ 0] + rh.e[ 1] * e[ 4] + rh.e[ 2] * e[ 8] + rh.e[ 3] * e[12];
	m.e[ 1] = rh.e[ 0] * e[ 1] + rh.e[ 1] * e[ 5] + rh.e[ 2] * e[ 9] + rh.e[ 3] * e[13];
	m.e[ 2] = rh.e[ 0] * e[ 2] + rh.e[ 1] * e[ 6] + rh.e[ 2] * e[10] + rh.e[ 3] * e[14];
	m.e[ 3] = rh.e[ 0] * e[ 3] + rh.e[ 1] * e[ 7] + rh.e[ 2] * e[11] + rh.e[ 3] * e[15];
	
	m.e[ 4] = rh.e[ 4] * e[ 0] + rh.e[ 5] * e[ 4] + rh.e[ 6] * e[ 8] + rh.e[ 7] * e[12];
	m.e[ 5] = rh.e[ 4] * e[ 1] + rh.e[ 5] * e[ 5] + rh.e[ 6] * e[ 9] + rh.e[ 7] * e[13];
	m.e[ 6] = rh.e[ 4] * e[ 2] + rh.e[ 5] * e[ 6] + rh.e[ 6] * e[10] + rh.e[ 7] * e[14];
	m.e[ 7] = rh.e[ 4] * e[ 3] + rh.e[ 5] * e[ 7] + rh.e[ 6] * e[11] + rh.e[ 7] * e[15];
	
	m.e[ 8] = rh.e[ 8] * e[ 0] + rh.e[ 9] * e[ 4] + rh.e[10] * e[ 8] + rh.e[11] * e[12];
	m.e[ 9] = rh.e[ 8] * e[ 1] + rh.e[ 9] * e[ 5] + rh.e[10] * e[ 9] + rh.e[11] * e[13];
	m.e[10] = rh.e[ 8] * e[ 2] + rh.e[ 9] * e[ 6] + rh.e[10] * e[10] + rh.e[11] * e[14];
	m.e[11] = rh.e[ 8] * e[ 3] + rh.e[ 9] * e[ 7] + rh.e[10] * e[11] + rh.e[11] * e[15];
	
	m.e[12] = rh.e[12] * e[ 0] + rh.e[13] * e[ 4] + rh.e[14] * e[ 8] + rh.e[15] * e[12];
	m.e[13] = rh.e[12] * e[ 1] + rh.e[13] * e[ 5] + rh.e[14] * e[ 9] + rh.e[15] * e[13];
	m.e[14] = rh.e[12] * e[ 2] + rh.e[13] * e[ 6] + rh.e[14] * e[10] + rh.e[15] * e[14];
	m.e[15] = rh.e[12] * e[ 3] + rh.e[13] * e[ 7] + rh.e[14] * e[11] + rh.e[15] * e[15];
	
	return m;
}

MMatrix4 MMatrix4::operator*( float s ) const {
	MMatrix4 m(*this);
	for(int i=0; i<16; ++i)
		m.e[i] *= s;
	return *this;
}

MMatrix4 MMatrix4::operator*=( const MMatrix4& rh ) {
	*this = *this * rh;
	return *this;
}

MMatrix4 MMatrix4::operator*=( float s ) {
	*this = *this * s;
	return *this;
}

float& MMatrix4::operator()(int row, int column) {
	return e[column+(row*4)];  //  e is stored in column major order
}

float& MMatrix4::operator[](int index) {
	return e[index];
}

const float& MMatrix4::operator()(int row, int column) const {
	return e[column+(row*4)];  //  e is stored in column major order
}

const float& MMatrix4::operator[](int index) const {
	return e[index];
}

void MMatrix4::transform( MVector &v ) const {
	MVector sv;
	sv[0] = v[0]*e[0]+v[1]*e[4]+v[2]*e[8]+e[12];
	sv[1] = v[0]*e[1]+v[1]*e[5]+v[2]*e[9]+e[13];
	sv[2] = v[0]*e[2]+v[1]*e[6]+v[2]*e[10]+e[14];
	v=sv;
}

void MMatrix4::translate( MVector &v ) const {
	v[0] += e[12];
	v[1] += e[13];
	v[2] += e[14];
}

void MMatrix4::rotate(MVector &v ) const {
	MVector r;

	r[0] = v[0]*e[0]+v[1]*e[4]+v[2]*e[8];
	r[1] = v[0]*e[1]+v[1]*e[5]+v[2]*e[9];
	r[2] = v[0]*e[2]+v[1]*e[6]+v[2]*e[10];
	
	v=r;
}

void MMatrix4::setTranslation( MVector v ) {
	e[12] = v[0];
	e[13] = v[1];
	e[14] = v[2];
}

MMatrix4 MMatrix4::transpose() const {
	MMatrix4 r(*this);
	std::swap(r.e[4],r.e[1]);
	std::swap(r.e[8],r.e[2]);
	std::swap(r.e[12],r.e[3]);
	std::swap(r.e[9],r.e[6]);
	std::swap(r.e[13],r.e[7]);
	std::swap(r.e[14],r.e[11]);
	return r;
}

MMatrix4 MMatrix4::scale(float x, float y, float z) const {
	MMatrix4 r(*this);
	r.e[0] *= x;
	r.e[5] *= y;
	r.e[10] *= z;
	return r;
}

void MMatrix4::output() const {
}

//  TODO: Implement non-affine inverse, clean up calculation(use direct indexes!)
MMatrix4 MMatrix4::inverse(bool affine) const {
	MMatrix4 r;
	if(affine) {
		float s = 1 / det();
		float v00 = ((*this)(1,1)*(*this)(2,2) - (*this)(1,2)*(*this)(2,1)) * s;
		float v01 = ((*this)(2,1)*(*this)(0,2) - (*this)(2,2)*(*this)(0,1)) * s;
		float v02 = ((*this)(0,1)*(*this)(1,2) - (*this)(0,2)*(*this)(1,1)) * s;
		float v10 = ((*this)(1,2)*(*this)(2,0) - (*this)(1,0)*(*this)(2,2)) * s;
		float v11 = ((*this)(2,2)*(*this)(0,0) - (*this)(2,0)*(*this)(0,2)) * s;
		float v12 = ((*this)(0,2)*(*this)(1,0) - (*this)(0,0)*(*this)(1,2)) * s;
		float v20 = ((*this)(1,0)*(*this)(2,1) - (*this)(1,1)*(*this)(2,0)) * s;
		float v21 = ((*this)(2,0)*(*this)(0,1) - (*this)(2,1)*(*this)(0,0)) * s;
		float v22 = ((*this)(0,0)*(*this)(1,1) - (*this)(0,1)*(*this)(1,0)) * s;
		float v30 = -(v00*(*this)(3,0) + v10*(*this)(3,1) + v20*(*this)(3,2));
		float v31 = -(v01*(*this)(3,0) + v11*(*this)(3,1) + v21*(*this)(3,2));
		float v32 = -(v02*(*this)(3,0) + v12*(*this)(3,1) + v22*(*this)(3,2));
		r(0,0) = v00;
		r(0,1) = v01;
		r(0,2) = v02;
		r(1,0) = v10;
		r(1,1) = v11;
		r(1,2) = v12;
		r(2,0) = v20;
		r(2,1) = v21;
		r(2,2) = v22;
		r(3,0) = v30;
		r(3,1) = v31;
		r(3,2) = v32;
	}
	else {
	}
	return r;
}

float MMatrix4::det() const {
	return	e[0] * e[5] * e[10] +
		e[4] * e[9] * e[2] +
		e[8] * e[1] * e[6] -
		e[8] * e[5] * e[2] -
		e[4] * e[1] * e[10] -
		e[0] * e[9] * e[6];
}

MMatrix4 makeTranslationMatrix(MVector pos) {
	MMatrix4 r;
	r.e[12] = pos.x();
	r.e[13] = pos.y();
	r.e[14] = pos.z();
	return r;
}

MMatrix4 eulerToMatrix( float x, float y, float z ) {
	MMatrix4 m;
	
	double cr = cos( x );
	double sr = sin( x );
	double cp = cos( y);
	double sp = sin( y );
	double cy = cos( z );
	double sy = sin( z );

	m.e[0] = ( float )( cp*cy );
	m.e[1] = ( float )( cp*sy );
	m.e[2] = ( float )( -sp );

	double srsp = sr*sp;
	double crsp = cr*sp;

	m.e[4] = ( float )( srsp*cy-cr*sy );
	m.e[5] = ( float )( srsp*sy+cr*cy );
	m.e[6] = ( float )( sr*cp );

	m.e[8] = ( float )( crsp*cy+sr*sy );
	m.e[9] = ( float )( crsp*sy-sr*cy );
	m.e[10] = ( float )( cr*cp );
	
	return m;
}

/*	Quaternion	*/
//  these are ALWAYS unit quaternions, which is why the data members are private and normal/normalize are private

Quaternion::Quaternion(float ax, float ay, float az, float qw) {
	x = ax;
	y = ay;
	z = az;
	w = qw;
}

Quaternion::Quaternion(float ax, float ay, float az) : x(ax), y(ay), z(az) {
	float term = 1.0f - (x*x) - (y*y) - (z*z);
	if (term < 0.0f) 
		w = 0.0f; 
	else 
		w = - (float) sqrt(term);
}

Quaternion::Quaternion(MMatrix4 matrix) {
	float trace = matrix[0] + matrix[5] + matrix[10] + matrix[15];
	float scale;
	if(trace > M_EPSILON)  {
		scale = 0.5 / sqrt(trace);
		w = 0.25 / scale;
		x = (matrix[6] - matrix[9]) * scale;
		y = (matrix[8] - matrix[2]) * scale;
		z = (matrix[1] - matrix[4]) * scale;
	}
	else if((matrix[0] > matrix[5]) && (matrix[0] > matrix[10])) { 
		scale = sqrt(matrix[15] + matrix[0] - matrix[5] - matrix[10]) * 2;
		w = (matrix[9] - matrix[6]) / scale;
		x = 0.25 * scale;
		y = (matrix[4] + matrix[1]) / scale;
		z = (matrix[8] + matrix[2]) / scale;
    }
	else if (matrix[5] > matrix[10]) {
		scale = sqrt(matrix[15] + matrix[5] - matrix[0] - matrix[10]) * 2;
		w = (matrix[8] - matrix[2]) / scale;
		x = (matrix[4] + matrix[1]) / scale;;
		y = 0.25 * scale;
		z = (matrix[9] + matrix[6]) / scale;
	}
	else { 
		scale = sqrt(matrix[15] + matrix[10] - matrix[0] - matrix[5]) * 2;
		w = (matrix[4] - matrix[1]) / scale;
		x = (matrix[8] + matrix[2]) / scale;
		y = (matrix[9] + matrix[6]) / scale;
		z = 0.25 * scale;
	}
}

MVector Quaternion::getAxis() const {
	float scale = sqrt(x*x + y*y + z*z);
	MVector toReturn;
	toReturn[0] = x / scale;
	toReturn[1] = y / scale;
	toReturn[2] = z / scale;
	return toReturn;
}

void Quaternion::normalize() {
	float magn = sqrt(w*w + x*x + y*y + z*z);
	w /= magn;
	x /= magn;
	y /= magn;
	z /= magn;
}

void Quaternion::invert() {
	*this = inverse();
}

Quaternion Quaternion::inverse() const {
	float magn = w*w + x*x + y*y + z*z;
	if(magn > 0.0) {
		magn = 1.0 / magn;
		float aw = w * magn;
		float ax = -x * magn;
		float ay = -y * magn;
		float az = -z * magn;
		return Quaternion(ax,ay,az,aw);
	}
	return *this;
}

Quaternion Quaternion::slerp(const Quaternion &q, float t) const
{
	float dot = w*q.w + x*q.x + y*q.y + z*q.z;
	int factor = 1;
	if(dot < 0.0)
		factor = -1;
		
	Quaternion result(	(1-t) * x + t * q.x * factor,
						(1-t) * y + t * q.y * factor,
						(1-t) * z + t * q.z * factor,
						(1-t) * w + t * q.w * factor );
						
	result.normalize();
	
	return result;
}

MMatrix4 Quaternion::getMatrix() const {
	MMatrix4 matrix;
	matrix[0] = 1 - 2*y*y - 2*z*z;
	matrix[1] = 2*x*y + 2*w*z;
	matrix[2] = 2*x*z - 2*w*y;

	matrix[4] = 2*x*y - 2*w*z;
	matrix[5] = 1 - 2*x*x - 2*z*z;
	matrix[6] = 2*y*z + 2*w*x;
  
	matrix[8] = 2*x*z + 2*w*y;
	matrix[9] = 2*y*z - 2*w*x;
	matrix[10] = 1 - 2*x*x - 2*y*y;
  
	matrix[3] = matrix[7] = matrix[11] = matrix[12] = matrix[13] = matrix[14] = 0.0;
	matrix[15] = 1.0;
	
	return matrix;
}

Quaternion Quaternion::operator*(const Quaternion &rhs) const {
	Quaternion result(	w*rhs.x + x*rhs.w + y*rhs.z - z*rhs.y,
						w*rhs.y + y*rhs.w + z*rhs.x - x*rhs.z,
						w*rhs.z + z*rhs.w + x*rhs.y - y*rhs.x,
						w*rhs.w - x*rhs.x - y*rhs.y - z*rhs.z );

	return result;
}

void Quaternion::rotate(MVector &v) const {
	Quaternion vec(v.x(), v.y(), v.z(), 0);

	Quaternion vec2 = (*this)*vec*inverse();
	v = MVector(vec2.x, vec2.y, vec2.z);
}

// Private part:

const float Quaternion::M_EPSILON = 0.0000001;

/*Quaternion eulerToQuaternion(float degrees, float x, float y, float z) {
	float halfTheta = M_PI * degrees / 360.0;
	float sinHalfTheta = sin(halfTheta);
	Quaternion	quat(	x * sinHalfTheta,
						y * sinHalfTheta,
						z * sinHalfTheta,
						cos(halfTheta) );
	quat.normalize();
	return quat;
}*/

Quaternion eulerToQuaternion(float heading, float attitude, float  bank) {
    // Assuming the angles are in radians.
    double c1 = cos(heading/2);
    double s1 = sin(heading/2);
    double c2 = cos(attitude/2);
    double s2 = sin(attitude/2);
    double c3 = cos(bank/2);
    double s3 = sin(bank/2);
    double c1c2 = c1*c2;
    double s1s2 = s1*s2;
	return Quaternion(	c1c2*s3 + s1s2*c3,
						s1*c2*c3 + c1*s2*s3,
						c1*s2*c3 - s1*c2*s3,
						c1c2*c3 - s1s2*s3 );
}