/************************************************************************************
3DMath grab bag: A bunch of classes and functions for working in 3D

All Code by Troy Lawlor Copyright (c)2005 All Rights Reserved.
*************************************************************************************/
 
#ifndef MATH_3DMATH_H
#define MATH_3DMATH_H

#include <math.h>
	
/*	simple types	*/

class MVertex2D {
	public:
	int x, y;
	MVertex2D() : x(0), y(0) {}
	MVertex2D(int p_x, int p_y) : x(p_x), y(p_y) {}
	void set(int p_x, int p_y) { x=p_x; y=p_y; }
};

class MVertexMemo2D : public MVertex2D {
	public:
	enum { NOT_TESTED=-1, OUT, IN };
	char result[2];
	
	inline void clearMemo() { result[0]=result[1]=NOT_TESTED; }
};

/*	3DMath classes	*/

class MMatrix4;

class MVector {
	public:
	float v[3];
	
	MVector();
	MVector( float p_x, float p_y, float p_z );
	MVector( float pv[] );
	MVector operator +(MVector) const;
	MVector operator -(MVector) const;
	MVector operator -() const;
	MVector operator *(float) const;
	float operator *(MVector) const;
	MVector operator *=(float);
	MVector operator +=(MVector);
	MVector operator -=(MVector);
	float& operator[](int index);
	const float& operator[](int index) const;
	float& x();
	const float& x() const;
	float& y();
	const float& y() const;
	float& z();
	const float& z() const;
	MVector normal() const;
	void normalize();
	float length() const;
};

class MPlane {
	public:
	MVector n;
	float d;
	
	MPlane();
	MPlane( MVector p_n, float p_d );
	MPlane( float p_a, float p_b, float p_c, float p_d );	
	void normalize();
	float operator *(MPlane) const;  //  dot-product
	MPlane operator *(float) const;
};

//  4x4 matrix stored in column-major order
class MMatrix4 {
	public:
		float e[16];

		MMatrix4();
		MMatrix4(float x, float y, float z, float xr, float yr, float zr);
		MMatrix4(float p_i[]);
		MMatrix4(MVector p);
		MMatrix4 operator*( const MMatrix4& rh ) const;
		MMatrix4 operator*( float s ) const;
		MMatrix4 operator*=( const MMatrix4& rh );
		MMatrix4 operator*=( float s );
		float& operator()(int row, int column);
		float& operator[](int index);
		const float& operator()(int row, int column) const;
		const float& operator[](int index) const;
		void transform( MVector &v ) const;
		void translate( MVector &v ) const;
		void rotate( MVector &v ) const;
		void setTranslation( MVector v );
		MMatrix4 scale(float x, float y, float z) const;
		MMatrix4 transpose() const;
		MMatrix4 inverse(bool affine=true) const;
		float det() const;
		void output() const;
};

class Quaternion
{
	public:
		Quaternion() : x(0), y(0), z(0), w(0) {}
		Quaternion(float p_x, float p_y, float p_z, float p_w);
		Quaternion(float p_x, float p_y, float p_z);
		Quaternion(MMatrix4 matrix);
  
		MMatrix4 getMatrix() const;
		MVector getAxis() const;
		void normalize();
		void invert();
		Quaternion inverse() const;
		Quaternion slerp(const Quaternion &q, float t) const;
		void rotate(MVector &v) const;
  
		Quaternion operator*(const Quaternion &rhs) const;
		
	public:
		float w, x, y, z;
		
	private:
		static const float M_EPSILON;
};

class MRay {
	public:
		MVector origin;
		MVector dir;
		
		MRay(MVector p_origin, MVector p_dir) : origin(p_origin), dir(p_dir) {}
};

//  misc. functions
inline int dist2(MVertex2D a, MVertex2D b) {
	return (a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y);
}

inline void decomposeAngle( float angle, float &rx, float &ry ) {
	rx = cos(angle);
	ry = sin(angle);
}

extern bool planeIntersection( const MPlane &a, const MPlane &b, const MPlane &c, MVector &result );

//  vector functions
extern float dotProduct( MVector a, MVector b );
extern MVector crossProduct( MVector a, MVector b );
extern float angleBetween( MVector a, MVector b );  //  NOTE: a and b must b unit vectors
inline float dist(MVector a, MVector b) {
	return sqrt((a[0]-b[0])*(a[0]-b[0]) + (a[1]-b[1])*(a[1]-b[1]) + (a[2]-b[2])*(a[2]-b[2]));
}
inline float dist2(MVector a, MVector b) {
	return ((a[0]-b[0])*(a[0]-b[0]) + (a[1]-b[1])*(a[1]-b[1]) + (a[2]-b[2])*(a[2]-b[2]));
}

//  matrix functions
extern MMatrix4 makeTranslationMatrix(MVector pos);
extern MMatrix4 eulerToMatrix( float x, float y, float z );

//  converts spherical floatinates to a vector(a unit vector if r is 1)
inline MVector sphericalToVector( float theta, float phi, float r=1.0f ) {
	MVector toReturn;
	toReturn[0] = r * sin(phi) * cos(theta);
	toReturn[1] = r * cos(phi);
	toReturn[2] = r * sin(theta) * sin(phi);
	return toReturn;
}

//  converts vector floatinates to spherical floatinates
inline void vectorToSpherical( MVector a, float &theta, float &phi, float &r ) {
	r = sqrt(a.v[0]*a.v[0] + a.v[1]*a.v[1] + a.v[2]*a.v[2]);
	theta = atan2(a.v[2],a.v[0]);
	phi = acos(a.v[1]/r);
}

//  quaternion functions
extern Quaternion eulerToQuaternion(float x, float y, float z);

#endif
